****************************************************************************************************
*                                           PROGRAM OVERVIEW
****************************************************************************************************
*
* PROGRAM: ms_deidentify.sas  
*
* Created (mm/dd/yyyy): 07/14/2014
* Last modified: 03/25/2016
* Version: 1.0
*
*--------------------------------------------------------------------------------------------------
* PURPOSE:
*  This program creates a de-identified dataset by assigning a randomly generated 'caseid' to one or 
*  more user-specified variables in an existing dataset. The program also produces crosswalk files for 
*  each variable replaced with a random 'caseid'. 
*  This program has utility in studies where individual-level data are needed but actual values of 
*  person-level identifying variables are required to be masked. The output dataset will contain all 
*  variables on the original dataset, with the randomly-generated 'caseid' variables replacing the 
*  user-specified variables. 
*
*
* PARAMETERS:
*  INFILE=     Name of the dataset containing variables to be masked
*  VARLIST=    Name(s) of variable(s) that requires replacement with a randomly-generated caseid
*  OUTFILE=    Name of the output dataset
*  XWALKLIB=   Libref to which crosswalk files will be saved
*
*--------------------------------------------------------------------------------------------------
* CONTACT INFO: 
*  Sentinel Coordinating Center
*  info@mini-sentinel.org
*
*--------------------------------------------------------------------------------------------------
*  CHANGE LOG:
*
*   Version   Date       Initials      Comment (reference external documentation when available)
*   -------   --------   --------   ---------------------------------------------------------------
*    0.2      08/20/15   MSOC (JRP)	Fixed typos.
*                                   Added in %abort code to stop processing and
*                                   issue custom WARNING if missing values found on any values 
*                                   for variables listing in VARLIST parameter.
*
*    1.0      11/20/15   MSOC (RR)  Restricted length of 'case_id' variables to maximum needed.
*                                   Explicitly declared local macro variables
***************************************************************************************************;

/* START ms_deidentify MACRO CODE */

%macro ms_deidentify(
  infile=,
  varlist=,
  outfile=,
  xwalklib=
  ); 

  %local varcount varlistsep miss i UniqueIDs Varlength j ;

  %put =====> MACRO CALLED: ms_deidentify v1.0;

  %*-----------------------
  * 1 - Process macro parameters.
  *-----------------------;
  %let varcount=%sysfunc(countw(&varlist.));

  %put =====>varlist=&varlist.;
  %put =====>varcount=&varcount.;
  %put =====>infile=&infile.;

  %*-----------------------
  * 2 - Check for missing values in any field listed in varlist parameter.
  *   - Abort program and issue custom WARNING to log if missing values found.
  *   - Program is not meant to handle replacement of missing fields.
  *-----------------------;
  %let varlistsep=%sysfunc(tranwrd(&varlist., %str( ), %str(,)));
  %put =====>varlistsep = &varlistsep.;

  data _chkmiss;
  		set &infile.;
  		if cmiss(&varlistsep.);
  run;
  data _null_;
  		if 0 then set _chkmiss nobs=n;
  		call symputx('miss',n);
  		stop;
  run;
  %put Number of obs with missing values across variables (&varlist.) is &miss.;
  %if %eval(&miss. GE 1) %then %do;
    %put Number of obs with missing values across variables (&varlist.) is &miss.;	
  	%put WARNING: One or more VARLIST variables contain missing values.;
  	%put WARNING: This program is not appropriate for use with replacement variables that contain missing values.;
  	%abort;
  %end;

  %*-----------------------
  * 3 - Create dataset of unique ids and a random number for each unique varlist variable.
  *-----------------------;
  %do i=1 %to &varcount.;
     proc sql noprint;
       create table _uniqueids_%scan(&varlist.,&i.) as
       select distinct %scan(&varlist.,&i.)
       from &infile.;

       create table _randomids_%scan(&varlist.,&i.) as
       select *, rand('uniform') as random
       from _uniqueids_%scan(&varlist.,&i.)
       order by random;

       %* Get number of unique IDs needed;
       select count(*) into :UniqueIDs
       from _uniqueids_%scan(&varlist.,&i.)
       ;
     quit;

  %*-----------------------
  * 4 - Create new 'caseid' for each variable in varlist.
  *-----------------------;
     %** Determine length of new 'caseid' variable = length of original variable name + # of digits to store max # of IDs;
     %let varlength=%eval(%length(%trim(%scan(&varlist.,&i.))) + %length(&UniqueIDs.));
     %put Varlength=&VarLength.;
     data &xwalklib..xwalk_%scan(&varlist.,&i.) (keep=%scan(&varlist.,&i.) %scan(&varlist.,&i.)_caseid);
        length %scan(&varlist.,&i.)_caseid $&VarLength.;
        set _randomids_%scan(&varlist.,&i.);
        %scan(&varlist.,&i.)_caseid="%scan(&varlist.,&i.)"||strip(put(_n_,20.));
     run;


  %*-----------------------
  * 5 - Merge 'caseid' variables and drop variables named in varlist.
  *   - Create final deidentifed dataset.
  *-----------------------;
     %let j = %eval(&varcount.+1);
     %put =====> j=&j.;
     %if &i.=1 %then %do;
        proc sql;
           create table _finaldataset_&i. as
           select a.*, b.%scan(&varlist.,&i.)_caseid from &infile. as a
             left join &xwalklib..xwalk_%scan(&varlist.,&i.) as b
             on a.%scan(&varlist.,&i.)=b.%scan(&varlist.,&i.);
        quit;
     %end;
     %do %until (&j. gt &i.);
        %if &i. gt 1 %then %do;
           proc sql;
              create table _finaldataset_&i. as
              select a.*, b.%scan(&varlist.,&i.)_caseid from _finaldataset_%eval(&i.-1) as a
                left join &xwalklib..xwalk_%scan(&varlist.,&i.) as b
                on a.%scan(&varlist.,&i.)=b.%scan(&varlist.,&i.);
           quit;
        %end;
        %if &i.=&varcount. %then %do;
           data &outfile.;
              set _finaldataset_&i. (drop=&varlist.);
           run;
        %end;
     %end;
  %end;


  %*-----------------------
  * 6 - Clean up.
  *-----------------------;
  proc datasets lib=work nolist nodetails;
    delete _chkmiss _uniqueids: _randomids: _final:;
  quit;

  %put NOTE: ********END OF MACRO: ms_deidentify v1.0********;

%mend ms_deidentify;

